/*
 * Copyright (C) 2023 Mattia Chiozzi <mattia@chiozzi.cc>
 * License: BSD0
 */

/*
 * motd - repeatedly print the message of the day on the screen
 *
 * The message of the day contains following information:
 * - the current date
 * - the current uptime
 * - the current CPU load
 * - the current CPU temperature
 * - the current memory usage
 * - the current disk usage
 * - the currently running processes
 * - the nameservers in use
 * - the last N notes (~/Document/notes) fitting on the terminal window
 *
 * ~ written at 04:21 AM while listening to Counterattack Mankind
 */

#include <kvm.h>
#include <limits.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/sensors.h>
#include <sys/statvfs.h>
#include <sys/sysctl.h>
#include <sys/ttycom.h>
#include <uvm/uvmexp.h>

/* get_net() */
#include <arpa/inet.h>   // for inet_ntop, INET6_ADDRSTRLEN, INET_ADDRSTRLEN
#include <net/if.h>      // for ifreq, ifr_addr, if_indextoname, IFNAMSIZ
#include <net/route.h>   // for rt_msghdr
#include <netinet/in.h>  // for sockaddr_in, INADDR_ANY, sockaddr_in6
#include <sys/ioctl.h>   // for ioctl
#include <sys/socket.h>  // for AF_INET, socket, AF_INET6, sockaddr, NET_RT_...
#include <sys/sockio.h>  // for SIOCGIFADDR

void motd(int sig);
void setup();
void cleanup(int sig);
void usage(void);
void die(const char *errstr);

void get_today(char *today);
void get_uptime(char *uptime);
void get_load(char *load);
void get_temp(char *temp);
void get_mem(char *mem);
void get_disk(char *disk);
void get_net(char *net);
void get_proc(char *proc);
void get_dns(char *dns);
void get_notes();

/* general */
#define BYTE_TO_GB		1073741824

/* colors */
#define GREEN	"\e[32m"	/* normal */
#define YELLOW	"\e[33m"	/* normal */
#define MAGENTA	"\e[1;35m"	/* bold */
#define RESET	"\e[0m"

/* limits */
#define SEC_MAX		99999	/* max seconds */
#define TODAY_MAX	11	/* 10 + \0 */
#define UPTIME_MAX	51	/* 51 + \0 */
#define LOAD_MAX	17	/* X.XX X.XX X.XX + \0 */
#define TEMP_MAX	12	/* XXX.XX degC + \0 */
#define MEM_MAX		99
#define DISK_MAX	35	/* XXX.X / XXXGB (XX%) + \0 */
#define NET_MAX		99	/* iface | ip + \0 */
#define PROC_MAX	30	/* X.. of which X.. is yours + \0 */
#define DNS_MAX		99
#define IPV6_MAX	46	/* 45 + \0 */

/* global variables */
char notes_path[PATH_MAX];

int
main(int argc, char *argv[])
{
	/* interval in seconds */
	int interval = 5;
	char dflag = 0;

	/* args */
	if (argc == 2) {
		if (argv[1][1] != 'd') usage();
		dflag = 1;
	} else if (argc == 3) {
		if (argv[1][1] != 's') usage();
		interval = strtonum(argv[2], 1, SEC_MAX, NULL);
	} else if (argc > 3)
		usage();

	/* build notes path */
	snprintf(notes_path, PATH_MAX, "%s/Documents/notes", getenv("HOME"));

	/* apply restricitons */
	setup();

	/* catch signals */
	signal(SIGINT, cleanup);
	signal(SIGWINCH, motd);

	/* hide cursor */
	printf("\e[?25l");

	while (1) {
		motd(0);
		if (dflag) cleanup(0);
		else sleep(interval);
	}

	return 0;
}

void
motd(int sig)
{
	char today[TODAY_MAX];
	char uptime[UPTIME_MAX];
	char load[LOAD_MAX];
	char temp[TEMP_MAX];
	char mem[MEM_MAX];
	char disk[DISK_MAX];
	char net[NET_MAX];
	char proc[PROC_MAX];
	char dns[DNS_MAX];

	get_today(today);
	get_uptime(uptime);
	get_load(load);
	get_temp(temp);
	get_mem(mem);
	get_net(net);
	get_disk(disk);
	get_proc(proc);
	get_dns(dns);

	printf(GREEN
	    "\e[1;1H\e[2J"
	    "██   ██ ███████ ██      ██       ████████\n"
	    "██   ██ ██      ██      ██       ██    ██\n"
	    "███████ █████   ██      ██       ██    ██\n"
	    "██   ██ ██      ██      ██       ██    ██\n"
	    "██   ██ ███████ ███████ ███████  ████████ ██\n"
	    "\n"
            MAGENTA"            ToDAY iS %s"RESET"\n"
	    "\n"
	    YELLOW"Uptime.......: "RESET"%s\n"
	    YELLOW"Load.........: "RESET"%s\n"
	    YELLOW"Temperature..: "RESET"%s\n"
	    YELLOW"Memory.......: "RESET"%s\n"
	    YELLOW"Disk Usage...: "RESET"%s\n"
	    YELLOW"Network......: "RESET"%s\n"
	    YELLOW"Processes....: "RESET"%s\n"
	    YELLOW"Nameservers..: "RESET"%s\n"
	    "\n",
	    today, uptime, load, temp, mem, disk, net, proc, dns
	    );

	get_notes();
}

void
setup()
{
	if (unveil("/etc/resolv.conf", "r") < 0)
		die("unveil");
	if (unveil(notes_path, "r") < 0)
		die("unveil");
	//if (pledge("stdio rpath tty vminfo", NULL) < 0)
		//die("pledge");
}

void
cleanup(int var)
{
	/* re-enable cursor */
	printf("\e[?25h");
	exit(EXIT_SUCCESS);
}

void
usage(void)
{
	fprintf(stderr, "  motd [-d | -s <seconds>]\n");
	exit(EXIT_FAILURE);
}

void
die(const char *errstr)
{
	/* re-enable cursor */
	printf("\e[?25h");
	perror(errstr);
	exit(EXIT_FAILURE);
}

void
get_today(char *today)
{
	time_t today_sec;
	struct tm *today_info;

	time(&today_sec);
	if ((today_info = localtime(&today_sec)) == NULL)
		die("localtime: err processing time\n");
	strftime(today, TODAY_MAX, "%d.%m.%Y", today_info);
}

void
get_uptime(char *uptime)
{
	struct timespec uptm_sec;
	int day, hrs, min;

	day = hrs = min = 0;
	if (clock_gettime(CLOCK_BOOTTIME, &uptm_sec) < 0)
		die("clock_gettime");

	if (uptm_sec.tv_sec >= 86400) {
		day = uptm_sec.tv_sec / 86400;
		uptm_sec.tv_sec %= 86400;
	}

	if (uptm_sec.tv_sec >= 3600) {
		hrs = uptm_sec.tv_sec / 3600;
		uptm_sec.tv_sec %= 3600;
	}

	if (uptm_sec.tv_sec >= 60) {
		min = uptm_sec.tv_sec / 60;
		uptm_sec.tv_sec %= 60;
	}

	if (day != 0)
		snprintf(uptime, UPTIME_MAX, "up %d day%s, %02d:%02d",
		    day, day > 1 ? "s" : "", hrs, min);
	else if (hrs != 0)
		snprintf(uptime, UPTIME_MAX, "up %02d:%02d", hrs, min);
	else
		snprintf(uptime, UPTIME_MAX, "up %d min%s",
		    min, min > 1 ? "s" : "");
}

void
get_load(char *load)
{
	double av[3];
	if (getloadavg(av, 3) < 0)
		die("getloadavg");
	snprintf(load, LOAD_MAX, "%.02f, %.02f, %.02f", av[0], av[1], av[2]);
}

void
get_temp(char *temp)
{
	struct sensordev sensordev;
	struct sensor sensor;
	size_t len;
	int dev, num;
	int mib[5] = {
		[0] = CTL_HW,
		[1] = HW_SENSORS,
		[3] = SENSOR_TEMP
	};

	for (dev = 0;; dev++) {
		mib[2] = dev;
		len = sizeof(sensordev);
		if (sysctl(mib, 3, &sensordev, &len, NULL, 0) < 0)
			break;
		len = sizeof(sensor);
		for (num = 0;; num++) {
			mib[4] = num;
			if (sysctl(mib, 5, &sensor, &len, NULL, 0) < 0)
				break;
			if (!strncmp(sensordev.xname, "cpu", 3)) {
				snprintf(temp, TEMP_MAX, "%.1f degC",
				    sensor.value * 1E-6 - 272.15);
				return;
			}
		}
	}

	/* no sensor detected */
	temp[0] = '-';
	temp[1] = '\0';
}

void
get_mem(char *mem)
{
	struct uvmexp uvm;
	int64_t tot_mem, used_mem;
	int mib[2], percent;
	size_t len;

	mib[0] = CTL_HW;
	mib[1] = HW_PHYSMEM64;
	len = sizeof(tot_mem);
	if (sysctl(mib, 2, &tot_mem, &len, NULL, 0) < 0)
		die("sysctl");

	mib[0] = CTL_VM;
	mib[1] = VM_UVMEXP;
	len = sizeof(uvm);
	if (sysctl(mib, 2, &uvm, &len, NULL, 0) < 0)
		die("sysctl");

	used_mem = (int64_t) uvm.active * uvm.pagesize;
	percent = used_mem * 100 / tot_mem;
	snprintf(mem, MEM_MAX, "%lld / %lldMB (%d%%)",
	    used_mem / 1048576, tot_mem / 1048576, percent);
}

void
get_disk(char *disk)
{
	struct statvfs buf;
	unsigned long long used, free, tot;
	int percent;

	if (statvfs("/home", &buf) < 0)
		die("sysctl");

	tot = (buf.f_blocks * buf.f_frsize) / BYTE_TO_GB;
	free = (buf.f_bfree * buf.f_frsize) / BYTE_TO_GB;
	used = tot - free;
	percent = used * 100 / tot;

	snprintf(disk, DISK_MAX, "%llu / %lluGB (%d%%)", used, tot, percent);
}

void
get_net(char *net)
{
	/* sysctl */
	int mib[6] = { CTL_NET, PF_ROUTE, 0, 0, NET_RT_DUMP, 0 };
	char *buf;
	size_t len;
	char ifname[IF_NAMESIZE] = { [0] = '\0' };
	struct rt_msghdr *rtm;
	struct sockaddr *sa;
	struct sockaddr_in *sin;

	/* get addr */
	int sockfd;
	struct ifreq ifr;
	char ip[INET6_ADDRSTRLEN];

	if (sysctl(mib, 6, NULL, &len, NULL, 0) < 0) die("sysctl");
	if (!(buf = malloc(len))) die("mallco");
	if (sysctl(mib, 6, buf, &len, NULL, 0) < 0) die("sysctl");

	for (rtm = (struct rt_msghdr *) buf;
	     (char *) rtm < buf + len;
	     rtm = (struct rt_msghdr *) ((char *) rtm + rtm->rtm_msglen)) {
		sin = (struct sockaddr_in *) (rtm + 1);

		/* stop at first found default interface */
		if (sin->sin_addr.s_addr == INADDR_ANY) {
			if_indextoname(rtm->rtm_index, ifname);
			break;
		}
	}

	if (ifname[0] == '\0') {
		net[0] = '-';
		net[1] = '\0';
		return;
	}

	if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
		die("socket");

	strlcpy(ifr.ifr_name, ifname, IFNAMSIZ - 1);
	ifr.ifr_addr.sa_family = AF_INET;

	if (ioctl(sockfd, SIOCGIFADDR, &ifr) == 0) {
		struct sockaddr_in *sa = (struct sockaddr_in *) & ifr.ifr_addr;
		inet_ntop(AF_INET, &(sa->sin_addr), ip, INET_ADDRSTRLEN);
	} else {
		ifr.ifr_addr.sa_family = AF_INET6;
		if (ioctl(sockfd, SIOCGIFADDR, &ifr) == 0) {
			struct sockaddr_in6 *sa =
			    (struct sockaddr_in6 *) & ifr.ifr_addr;
			inet_ntop(AF_INET6, &(sa->sin6_addr), ip,
			    INET6_ADDRSTRLEN);
		} else {
			fprintf(stderr, "err: Unable to retrieve address\n");
			return;
		}
	}

	snprintf(net, NET_MAX, "%s | %s", ifname, ip);

	close(sockfd);
	free(buf);
}

void
get_proc(char *proc)
{
	kvm_t *kd;
	struct kinfo_proc *procs;
	int p_tot, p_usr;

	if (!(kd = kvm_openfiles(NULL, NULL, NULL, KVM_NO_FILES, NULL)))
		die("kvm_openfiles");

	if(!(procs = kvm_getprocs(kd, KERN_PROC_ALL, 0,
	    sizeof(struct kinfo_proc), &p_tot)))
		die("kvm_getprocs");

	if (!(procs = kvm_getprocs(kd, KERN_PROC_UID, getuid(),
	    sizeof(struct kinfo_proc), &p_usr)))
		die("kvm_getprocs");

	snprintf(proc, PROC_MAX, "%d of which %d is yours", p_tot, p_usr);
	kvm_close(kd);
}

void
get_dns(char *dns)
{
	FILE *resolv = fopen("/etc/resolv.conf", "r");
	char *line = NULL, tmp[IPV6_MAX];
	size_t len = 0;

	dns[0] = '\0';
	if (!resolv)
		return;

	while (getline(&line, &len, resolv) != EOF) {
		sscanf(line, "nameserver %s", tmp);
		if (dns[0] == '\0')
			strlcpy(dns, tmp, DNS_MAX);
		else if (tmp[0] != '\0')
			snprintf(dns, DNS_MAX, "%s, %s", dns, tmp);
		tmp[0] = '\0';
	}

	fclose(resolv);
	if (line)
		free(line);
}

void
get_notes()
{
	FILE *notes;
	struct winsize win;
	char *line = NULL;
	size_t len = 0;
	int i = 1, ch, lines_avail, lines_tot = 0;

	ioctl(STDOUT_FILENO, TIOCGWINSZ, &win);
	if (!(notes = fopen(notes_path, "r")))
		return;

	/* count tot lines */
	while ((ch = fgetc(notes)) != EOF)
		if (ch == '\n')
			lines_tot++;

	/* calc lines available & total note lines */
	lines_avail = win.ws_row - 18;
	if (lines_avail <=0) return;
	lines_tot -= lines_avail;
	rewind(notes);

	/* skip lines */
	while (getline(&line, &len, notes) != EOF && i++ < lines_tot);

	/* print last N lines */
	while (getline(&line, &len, notes) != EOF) {
		if (strlen(line) > 39) {
			line[39] = '.';
			line[40] = '.';
			line[41] = '.';
			line[42] = '\n';
			line[43] = '\0';
		}
		printf(" %s", line);
	}
	
	fclose(notes);
	if (line)
		free(line);
}
